/* vim: set sw=8 ts=8 si : */
/*********************************************
* Author: Guido Socher, Copyright: GPL 
* Clock frequency     : 4,000000 MHz
* Copyright: GPL
**********************************************/
#include <avr/signal.h>
#include <string.h>
#include <avr/io.h>
#include "uart.h"
#include "avr_compat.h"

/* the RTSOUT must be on port D */
#define RTSOUT PD2

/* define either atmeag8 or at4433 not both !! */
#ifdef UCSRC
#define ATMEGA8 1
#else
#define AT4433 1
#endif

static char uart_rxbuf[UART_RX_BUFFER_SIZE+1];
volatile static unsigned char uart_rx_pos; /* end of string in rx */
volatile static unsigned char uart_rx_linecomplete; /* one complete line in buffer */
volatile static unsigned char uart_tx_busy; /* uart is still transmitting */

/* the following function will be called when the receive is complete 
 * SIGNAL =interrupts are disabled */
SIGNAL(SIG_UART_RECV) {
	uart_rxbuf[uart_rx_pos]=inp(UDR);
	if (uart_rxbuf[uart_rx_pos] == '\n' || uart_rxbuf[uart_rx_pos] == '\r'){
		uart_rx_linecomplete=1;
	}else{
		uart_rx_pos++;
		/* prevent overflow */
		if (uart_rx_pos > UART_RX_BUFFER_SIZE){
			uart_rx_pos=0;
		}
		if (uart_rx_pos > UART_RX_BUFFER_SIZE -1 ){
			/* buffer full, set RTS/CTS to off */
			uart_rx_linecomplete=1;
			cbi(PORTD,RTSOUT);
		}
	}
}

/* the following function will be called when the data
* in UDR register is sent out, INTERRUPT=interrupts are not disabled */
//==== SIG_UART_DATA
INTERRUPT(SIG_UART_TRANS) {
	uart_tx_busy=0;
}


void uart_init(void) 
{
	uart_tx_busy=0;
	uart_rx_pos=0;
	uart_rx_linecomplete=0;

	#ifdef AT4433
	outp((unsigned char) 25, UBRR); /* set baud rate for 4MHz, 25=9600, 12=19200 */
	outp(BV(RXCIE)|BV(RXEN)|BV(TXCIE)|BV(TXEN),UCSRB); /* enable tx/rx and interrupt on tx/rx */
	#endif

	#ifdef ATMEGA8
	outp((unsigned char)  0, UBRRH);
	outp((unsigned char) 25, UBRRL); /* set baud rate for 4MHz, 25=9600, 12=19200 */
	outp(BV(RXCIE)|BV(RXEN)|BV(TXCIE)|BV(TXEN),UCSRB); /* enable tx/rx and interrupt on tx/rx */
	/* format: asynchronous, 8data, no parity, 1stop bit */
        UCSRC = (1<<URSEL)|(3<<UCSZ0);
	#endif

	/* configure the RTS as output */
	sbi(DDRD,RTSOUT);
	/* RTS on = I am ready to receive data */
	sbi(PORTD,RTSOUT);

	/* you must enable interrupt in main prog by calling sei() */

}
void uart_reset(void)
{
	uart_tx_busy=0;
	uart_rx_pos=0;
	uart_rx_linecomplete=0;
	sbi(PORTD,RTSOUT);
}
/* send one character to the rs232 */
void uart_sendchar(char c) 
{
	while (uart_tx_busy);
	uart_tx_busy=1;
	outp(c, UDR);
}
/* send string to the rs232 */
void uart_sendstr(char *s) 
{
	while (*s){
		uart_sendchar(*s);
		s++;
	}
}

void uart_sendstr_p(const prog_char *progmem_s)
/* print string from program memory on rs232 */
{
	char c;
	while ((c = pgm_read_byte(progmem_s++))) {
		uart_sendchar(c);
	}

}


/* read a complet line from the rs232 */
unsigned char uart_getrxbufnl(char *uart_outbuf)  
{
	if (uart_rx_linecomplete==0) return(0);
	uart_rxbuf[uart_rx_pos]='\0';
	strcpy(uart_outbuf,uart_rxbuf);
	/* RTS on = I am ready to receive data */
	sbi(PORTD,RTSOUT);
	uart_rx_pos=0;
	uart_rx_linecomplete=0;
	return(1);
}

